# -*- coding: utf-8 -*-
"""
fddl.py
fdtd2d/fdtd3d対応
"""

import sys
import numpy as np
import torch
from torch import nn, optim
from torchvision import models
import matplotlib.pyplot as plt
#from torchinfo import summary

import utils, dl, fit
import myCNN, myResNet
import plotutils

def main(argv):
    # FDTD計算結果ファイル名(誘電率とS行列からなる)
    #fdtd_bin = '../fdtd2d/fdtd2d.bin'
    fdtd_bin = '../fdtd3d/fdtd3d.bin'

    # モデルファイル
    modelfile = 'fddl.pth'
    load_model = 0     # 通常0, 前回保存したmodelfileからrestartするとき1
    save_model = 1     # 通常1, 計算終了時にmodelfileを保存しないとき0

    # 計算パラメーター
    ndata = -1         # データ数, -1のときはすべてのデータ
    batch_size = 50    # バッチサイズ, 通常50-100程度
    num_epochs = 50    # エポック数
    component = 1      # S行列成分, 1=S, 2=Re(S),Im(S), 3=Re(S),Im(S),S
    train_ratio = 0.8  # 訓練データの割合(通常0.8程度, テストのときは0)
    mixed = True and torch.cuda.is_available()  # 混合精度

    # device
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

    # 乱数初期化
    utils.torch_seed()

    # FDTD計算結果を読み込む(fdtd2dではNz=1)
    Ndata, Nx, Ny, Nz, NTx, NRx, E, S = utils.readdata(fdtd_bin)

    # 図形確認(debug)
    #_plot2d_data(E, S); return

    # dataset
    train_set, test_set = dl.dataset(ndata, component, E, S, train_ratio)

    # dataloader
    train_loader, test_loader = dl.dataloader(train_set, test_set, batch_size=batch_size)
    #image, label = next(iter(train_loader))
    #print(image.shape, label.shape); return

    # 出力サイズ
    out_size = Nx * Ny * Nz

    # myCNN/myResMet
    in_size = (component, NTx, NRx)
    #model = myCNN.CNN10(in_size, [256], [0], out_size)
    model = myResNet.ResNet(in_size[0], 5, 256, 0, out_size)

    # ResNet(momentum?, 性能悪い?)
    #model = models.resnet18(weights=models.ResNet18_Weights.DEFAULT)
    #model = models.resnet50(weights=models.ResNet50_Weights.DEFAULT)
    #model.conv1 = nn.Conv2d(component, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)
    #model.fc = nn.Linear(model.fc.in_features, out_size)

    # VGG(component=3のみ, 32px必要,momentum?, 性能悪い?)
    #model = models.vgg11_bn(weights=None)
    #model = models.vgg11_bn(weights=models.VGG11_BN_Weights.DEFAULT)
    #model.classifier[6] = nn.Linear(model.classifier[6].in_features, out_size)
    #summary(model, (50, 1, 16, 16)); return
    #print(model); return

    # restart（前回と同じmodelのとき有効）
    if load_model == 1:
        utils.load_model(model, modelfile, device)

    # 損失関数
    criterion = nn.MSELoss()

    # 最適化関数
    #optimizer = optim.SGD(model.parameters(), lr=0.001, momentum=0.9)
    optimizer = optim.Adam(model.parameters())

    # 計算条件表示
    print('train=%d test=%d %dx%dx%d %dx%dx%d=%d epochs=%d'
        % (len(train_set), len(test_set), NTx, NRx, component, Nx, Ny, Nz, Nx * Ny * Nz, num_epochs))

    # 学習実行(計算時間の主要部)
    preds, history = fit.fit(model, device, num_epochs, train_loader, test_loader, optimizer, criterion, mixed,
        modelfile if (save_model == 1) else None)

    # 計算結果をファイルに保存する
    datafile = 'fddl.npz'
    utils.save_data(datafile, Nx, Ny, Nz, test_set, preds)

    # debug, loss履歴を図形表示する
    _plot2d_history(history)

# debug, 入力データを図形表示する
def _plot2d_data(D, S):
    # 誘電率分布
    #print(D.shape, np.min(D), np.max(D))
    if D.ndim == 3:
        # fdtd2d: 指定したデータ
        f = D
        figtitle = '誘電率(fdtd2d)'
    elif D.ndim == 4:
        # fdtd3d: 指定したZ面の指定したデータ
        k = 10
        k = max(0, min(D.shape[3] - 1, k))
        f = D[:, :, :, k]
        figtitle = '誘電率(fdtd3d Z=%d)' % k
    #print(np.min(D), np.max(D))
    sta = [str(i + 1) for i in range(D.shape[0])]
    plotutils.contour((8, 6), 4, 4, 1, f, np.min(D), np.max(D), sta, figtitle)

    # S行列: 指定したデータ
    #print(np.min(S), np.max(S))
    sta = [str(i + 1) for i in range(S.shape[0])]
    s_r = S[:, 0, :, :]
    s_i = S[:, 1, :, :]
    s_a = np.sqrt(s_r**2 + s_i**2)
    #plotutils.contour((8, 6), 12, 8, 10, s_r, np.min(s_r), np.max(s_r), sta, 'Re(S)')
    #plotutils.contour((8, 6), 12, 8, 10, s_i, np.min(s_i), np.max(s_i), sta, 'Im(S)')
    #plotutils.contour((8, 6), 12, 8, 10, s_a, np.min(s_a), np.max(s_a), sta, 'S')

# debug, loss履歴を図形表示する
def _plot2d_history(history):
    if len(history) < 2:
        return
    plt.figure('loss', figsize=(6, 4), layout='tight')
    plt.plot(history)
    plt.xlim(0, len(history) - 1)
    plt.ylim(0, 0.3)
    plt.grid(True)
    plt.xlabel('epoch')
    plt.ylabel('loss')
    plt.show()

if __name__ == "__main__":
    main(sys.argv)
