# -*- coding: utf-8 -*-
"""
plot.py
計算結果を図形表示する
"""

import os, datetime
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

import utils, plotutils

def main():
    
    # テストデータのラベルと予測値をファイルから読み込む
    datafile = 'fddl.npz'
    Nx, Ny, Nz, labels, preds = utils.load_data(datafile)
    #print(Nx, Ny, Nz, labels.shape, preds.shape)

    # ファイル更新日時
    dt = datetime.datetime.fromtimestamp(os.path.getmtime(datafile)).strftime('%Y/%m/%d %H:%M:%S')

    # 1D->4D配列
    labels = labels.reshape(-1, Nx, Ny, Nz)
    preds  = preds.reshape(-1, Nx, Ny, Nz)
    #print(labels.shape, preds.shape)
    assert labels.shape == preds.shape

    # 等高線図(fdtd2d/fdtd3d)
    _plot2d_labels((7, 5), 3, 4, labels, preds, dt)

    # 線上分布図(fdtd2d)
    """
    if Nz == 1:
        _plot2d_profile('profile', labels, preds)
    """
    
    # 3D分布図(fdtd3d)
    """
    if Nz > 1:
        idata = 300  # 適当に代入する
        idata = max(0, min(idata, labels.shape[0] - 1))
        vmin = min(np.min(labels[idata]), np.min(preds[idata]))
        vmax = max(np.max(labels[idata]), np.max(preds[idata]))
        _plot3d('label data#=%d' % (idata + 1), labels[idata], vmin, vmax)
        _plot3d('pred data#=%d' % (idata + 1), preds[idata], vmin, vmax)
    """

# 2D画像
def _plot2d_labels(figsize, mx, my, labels, preds, dt):
    # 最小・最大
    vmin = min(np.min(labels), np.min(preds))
    vmax = max(np.max(labels), np.max(preds))
    #print(vmin, vmax)

    # 図形表示
    if labels.shape[3] == 1:
        # fdtd2d
        figtitle = 'fdtd2d label-pred %s' % dt
        p1 = labels[:,:,:,0]  # 4D->3D
        p2 = preds[:,:,:,0]
        mstep = int(np.ceil(p1.shape[0] / (mx * my)))
        # 平均, 誤差
        ave = labels.mean(axis=(1,2,3))
        dif = np.abs(labels - preds).mean(axis=(1,2,3))
        str_ave = ['%d %.4f' % (i + 1, d) for i, d in enumerate(ave)]
        str_dif = ['%d %.4f' % (i + 1, d) for i, d in enumerate(dif)]
        # 図形出力
        #plotutils.contour(figsize, mx, my, mstep, p1, vmin, vmax, str_ave, 'label')
        #plotutils.contour(figsize, mx, my, mstep, p2, vmin, vmax, str_dif, 'pred')  # 推論
        plotutils.contour2(figsize, mx, my, mstep, p1, p2, vmin, vmax, str_ave, str_dif, figtitle)
    else:
        # fdtd3d
        n = 5  # 適当に代入する
        n = max(0, min(n, labels.shape[0] - 1))
        figtitle = 'fdtd2d label-pred data#=%d %s' % (n + 1, dt)
        p1 = labels[n].transpose(2,0,1)  # XYZをZXYの順に変換する
        p2 = preds[n].transpose(2,0,1)
        mstep = int(np.ceil(p1.shape[0] / (mx * my)))
        # 平均, 誤差
        ave = p1.mean(axis=(1,2))
        dif = np.abs(p1 - p2).mean(axis=(1,2))
        str_ave = ['%d %.4f' % (i + 1, d) for i, d in enumerate(ave)]
        str_dif = ['%d %.4f' % (i + 1, d) for i, d in enumerate(dif)]
        # 図形出力
        #plotutils.contour(figsize, mx, my, mstep, p1, vmin, vmax, str_ave, 'label')
        #plotutils.contour(figsize, mx, my, mstep, p2, vmin, vmax, str_dif, 'pred')  # 推論
        plotutils.contour2(figsize, mx, my, mstep, p1, p2, vmin, vmax, str_ave, str_dif, figtitle)

# 3D分布図
def _plot3d(figtitle, a3d, vmin, vmax):
    nx, ny, nz = a3d.shape
    #print(vmin, vmax)
    #print(nx, ny, nz)

    verts = []
    colors = []
    for k in range(nz):
        for i in range(nx):
            for j in range(ny):
                rect = [[i, j, k], [i + 1, j, k], [i + 1, j + 1, k], [i, j + 1, k], [i, j, k]]
                verts.append(rect)
                v = (a3d[i, j, k] - vmin) / (vmax - vmin)
                color = (1, 1 - v, 1 - v)
                colors.append(color)
    verts = np.array(verts)
    colors = np.array(colors)

    fig = plt.figure(figtitle, figsize=(6, 5), layout='tight')
    ax = fig.add_subplot(projection='3d')
    poly = Poly3DCollection(verts, facecolor=colors, alpha=0.3)
    ax.add_collection3d(poly)

    ax.set_xlim(0, nx)
    ax.set_ylim(0, ny)
    ax.set_zlim(0, nz)
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.set_aspect('equal')
    ax.view_init(elev=20, azim=225, roll=0)
    ax.grid(visible=False)

    suptitle = 'Nx=%d Ny=%d Nz=%d min=%.5f max=%.5f' % (nx, ny, nz, vmin, vmax)
    plt.suptitle(suptitle)

    plt.show()

# 線上分布図(2D)
def _plot2d_profile(figtitle, labels, preds):
    fig = plt.figure(figtitle, figsize=(8, 6), layout='tight')
    ax = fig.add_subplot()

    p1 = labels[:,:,:,0]  # 4D->3D
    p2 = preds[:,:,:,0]
    #print(p1.shape, p2.shape)

    idata = 0 #preds.shape[0] - 1
    for i in range(p1.shape[2]):
        ax.plot(p1[idata, :, i], color='r')
        ax.plot(p2[idata, :, i], color='k')

    ax.set_xlim(0, p1.shape[2] - 1)
    ax.set_ylim(0.5, 3.0)

    plt.show()

if __name__ == "__main__":
    main()
