/*
FDTD (2D)
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <time.h>

#ifdef _OPENMP
#include <omp.h>
#endif

#include "alloc.h"

extern void make_mesh(int, int, float, float, float, float *, float *, float *, float *, float *, float *, float *);
extern void txrx(int, int, int, int, int, int, int *, int *, int, int, int *, int *);
extern void geometry(int, int, int64_t, const int [], const float [], const float [], int, int, int, int, float ***, float ***, float, float, int **);
extern void solve(int, int, int, float, float, int, int, int, int, int *, int *, float **, float **, float **, float **, float **, float *, float *, float *, float *, float *, float *, float **, float **, float, float, float **, float **, float **, float **, float *, float *, int, int, float **, float **, float **);
extern void writedata(const char [], int, int, int, int, int, int, float ***, float ***, float ***);
extern void plot2d_g(int, int, int, int, int, int, int, float ***, int, int, int, int *, int *, int, int *, int *, const char []);
extern void plot2d_e(int, int, int, int, int, int, int, float ****, int, int, int *, int *, int, int **, const char []);
extern void plot2d_s(int, int, int, int, int, int, float ***, float ***, int, const char []);
extern void log_s1(int, int, int, float ***, float ***);
extern void log_s2(int, int, int, float ***, float ***);
extern int memory_size(int, int, int, int, int, int, int, int);
extern int file_size(int, int, int, int, int, int);
extern void mpi_init(int, char *[], int *, int *);
extern void mpi_close(void);
extern void comm_S(int, int, int, int, int, int, float ***, float ***);
extern double cputime(void);

const float C = 2.99792458e8f;  // 光速

int main(int argc, char *argv[])
{
	// 定数
	const int   Ndata = 1000;           // データ数
	const int   Nrect = 3;               // 長方形数
	const int   Lrect[] = {3, 7, 1};     // 長方形パラメーター(最小,最大,余白)
	const float Eps[] = {1.5f, 2.5f};    // 誘電率(最小,最大)
	const float Sig[] = {0.02f, 0.08f};  // 導電率[S/m](最小,最大)
	const int   Nsmot = 2;               // 平滑化回数(0以上)
	const int   NTx   = 16;              // 送信点数
	const int   NRx   = 16;              // 受信点数
	const int   DTx   = 1;               // 送信点間隔セル数
	const int   DRx   = 1;               // 受信点間隔セル数
	const float Dx    = 0.005f;          // X方向セルサイズ
	const float Dy    = 0.005f;          // Y方向セルサイズ
	const int   Nout  = 4;               // 外部領域セル数
	const int   Lant  = 2;               // アンテナ位置(対象領域境界からの距離)
	const int   ABC   = 6;               // PML層数(0のときはMur一次)
	const float Freq  = 3e9f;            // 周波数
	const int   Niter = 600;             // タイムステップ数
	const float Eps0  = 1;               // 背景媒質比誘電率
	const float Sig0  = 0;               // 背景媒質導電率[S/m]
	const int   My    = (NTx + 1) * DTx;
	const int   Mx    = My;
	const int   Nx    = Mx + (2 * Nout);
	const int   Ny    = My + (2 * Nout);
	const int64_t Seed = 10000;          // 乱数の種, 10000:学習用, 20000:テスト用
	char outfile[BUFSIZ] = "fdtd.bin";   // 出力ファイル名

	// debug
	const int log1 = 1;  // (数値出力) 計算条件とS行列平均(0/1, 通常1)
	const int log2 = 0;  // (数値出力) 収束状況(0/1, 通常0)
	const int log3 = 0;  // (数値出力) S行列(0/1, 通常0)
	const int fig1 = 0;  // (図形出力) 誘電率/導電率分布(0/1, 通常0)
	const int fig2 = 0;  // (図形出力) 電界分布(0,1,2..., 複数ページ対応, 通常0)
	const int fig3 = 0;  // (図形出力) S行列(0/1, 通常0)
	const char fig1ev2[] = "fig1.ev2";  // $ ev2d.exe fig1.ev2
	const char fig2ev2[] = "fig2.ev2";  // $ ev2d.exe fig2.ev2
	const char fig3ev2[] = "fig3.ev2";  // $ ev2d.exe fig3.ev2

	// MPI
	int commsize, commrank;
	mpi_init(argc, argv, &commsize, &commrank);
	//printf("%d %d\n", commsize, commrank); fflush(stdout);

	// 引数: スレッド数、出力ファイル
	int nthread = 1;
	if (argc > 1) {
		nthread = atoi(argv[1]);
	}
	if (argc > 2) {
		strcpy(outfile, argv[2]);
	}

	// OpenMP
#ifdef _OPENMP
	omp_set_num_threads(nthread);
#endif

	assert((Nx > 0) && (Ny > 0) && (Mx > 0) && (My > 0));
	assert((NTx > 0) && (NRx > 0) && (DTx > 0) && (DRx > 0));
	assert(Nout > Lant);
	assert(ABC >= 0);

	// 開始時刻
	const double tstart = cputime();

	// メッシュ作成
	float Dt;
	float *Xn, *Yn, *dXn, *dYn, *dXc, *dYc;
	alloc1d(float, Xn,  Nx + 1)
	alloc1d(float, Yn,  Ny + 1)
	alloc1d(float, dXn, Nx + 1)
	alloc1d(float, dYn, Ny + 1)
	alloc1d(float, dXc, Nx + 0)
	alloc1d(float, dYc, Ny + 0)
	make_mesh(Nx, Ny, Dx, Dy, Eps0, &Dt, Xn, Yn, dXn, dYn, dXc, dYc);

	// 送受信点座標
	int *iTx, *jTx, *iRx, *jRx;
	alloc1d(int, iTx, NTx)
	alloc1d(int, jTx, NTx)
	alloc1d(int, iRx, NRx)
	alloc1d(int, jRx, NRx)
	txrx(Nx, Ny, Nout, Lant, NTx, DTx, iTx, jTx, NRx, DRx, iRx, jRx);

	// 計算条件表示
	if (log1 && (commrank == 0)) {
		const int mem = memory_size(nthread, Ndata, Nx, Ny, NTx, NRx, ABC, fig2);
		const int filesize = file_size(Ndata, Mx, My, 1, NTx, NRx);
		printf("data=%d Nx=%d Ny=%d Mx=%d My=%d Tx=%d Rx=%d rect=%d(%d)\n",
			Ndata, Nx, Ny, Mx, My, NTx, NRx, Nrect, Nsmot);
		printf("process=%d thread=%d memory=%dMB ABC=%d iter=%d output=%s(%dMB)\n",
			commsize, nthread, mem, ABC, Niter, outfile, filesize);
		fflush(stdout);
	}

	// 長方形(debug)
	int **rect;
	alloc2d(int, rect, Ndata, 4 * Nrect)

	// 誘電率配列(セル中心, xデータ数)
	float ***Epsr, ***Sigm;
	alloc3d(float, Epsr, Ndata, Nx, Ny)
	alloc3d(float, Sigm, Ndata, Nx, Ny)

	// 誘電率配列代入
	geometry(Ndata, Nrect, Seed, Lrect, Eps, Sig, Nsmot, Nx, Ny, Nout, Epsr, Sigm, Eps0, Sig0, rect);

	// debug, 全領域誘電率分布図
	if (((fig1 == 1) || (fig1 == -1)) && (commrank == 0)) {
		plot2d_g(12, 8, Ndata, 1, Nx, Ny, Nout, Epsr, 1, 0, NTx, iTx, jTx, NRx, iRx, jRx, fig1ev2);
	}
	if (fig1 < 0) {
		exit(0);  // debug: 計算せずに終了する
	}

	// debug, 全領域電界絶対値, 図形出力用, セル中心, データ数x送信点の配列
	float ****Ea = NULL;
	if (fig2 && (commrank == 0)) {
		alloc4d(float, Ea, fig2, NTx, Nx, Ny)
	}

	// S行列(xデータ数)
	float ***S_r = NULL, ***S_i = NULL;
	alloc3d(float, S_r, Ndata, NTx, NRx)
	alloc3d(float, S_i, Ndata, NTx, NRx)

	// プロセスごとのデータ番号の範囲（ブロック分割）
	const int ndata = (Ndata + (commsize - 1)) / commsize;
	int idata0 = commrank * ndata;
	if (idata0 > Ndata - 1) {
		idata0 = Ndata - 1;
	}
	int idata1 = idata0 + ndata;
	if (idata1 > Ndata) {
		idata1 = Ndata;
	}
	//printf("%d %d %d %d %d %d\n", commsize, commrank, Ndata, ndata, idata0, idata1); fflush(stdout);

	// OpenMP並列
#ifdef _OPENMP
#pragma omp parallel
#endif
{
	// FDTD係数(スレッド変数)
	float **C1Ez = NULL, **C2Ez = NULL;
	alloc2d(float, C1Ez, Nx + 1, Ny + 1)
	alloc2d(float, C2Ez, Nx + 1, Ny + 1)

	// ABC変数(スレッド変数)
	float **Ezmurx = NULL, **Ezmury = NULL;
	float **Ezx = NULL, **Ezy = NULL;
	if (ABC == 0) {
		alloc2d(float, Ezmurx, 2, Ny + 1)
		alloc2d(float, Ezmury, 2, Nx + 1)
	}
	else {
		alloc2d(float, Ezx, Nx + 1 + 2 * ABC, Ny + 1 + 2 * ABC)
		alloc2d(float, Ezy, Nx + 1 + 2 * ABC, Ny + 1 + 2 * ABC)
	}

	// 電磁界配列(スレッド変数)
	float **Ez = NULL, **Hx = NULL, **Hy = NULL;
	alloc2d(float, Ez, Nx + 1 + 2 * ABC, Ny + 1 + 2 * ABC)
	alloc2d(float, Hx, Nx + 1 + 2 * ABC, Ny + 0 + 2 * ABC)
	alloc2d(float, Hy, Nx + 0 + 2 * ABC, Ny + 1 + 2 * ABC)

	// debug, 全領域調和界(スレッド変数, 節点)
	float **Ez_r = NULL, **Ez_i = NULL;
	if (fig2 && (commrank == 0)) {
		alloc2d(float, Ez_r, Nx + 1, Ny + 1)
		alloc2d(float, Ez_i, Nx + 1, Ny + 1)
	}

	// スレッド番号
	const int ithread = omp_get_thread_num();

	// データに関するループ
	for (int idata = idata0 + ithread; idata < idata1; idata += nthread) {
		const int ilog = log2 && (commrank == 0);  // 収束状況
		const int iout = (idata < fig2) && (commrank == 0);  // 電界分布
		if (log1 && (commrank == 0) && (idata % 1000 == 0)) {printf("%d %d\n", idata, ithread); fflush(stdout);}  // 経過出力
		for (int itx = 0; itx < NTx; itx++) {
			solve(
				Niter, Nx, Ny, Freq, Dt, ABC,
				iTx[itx], jTx[itx], NRx, iRx, jRx,
				Ez, Hx, Hy, Ezx, Ezy,
				Xn, Yn, dXn, dYn, dXc, dYc,
				Epsr[idata], Sigm[idata], Eps0, Sig0,
				C1Ez, C2Ez,
				Ezmurx, Ezmury,
				S_r[idata][itx], S_i[idata][itx],
				ilog, iout, Ez_r, Ez_i, (iout ? Ea[idata][itx] : NULL));
		}
	}
}
	// S行列をrootに集める
	if (commsize > 1) {
		comm_S(commsize, commrank, idata0, idata1, NTx, NRx, S_r, S_i);
	}

	// 出力:root
	if (commrank == 0) {
		// ファイル出力
		// ファイルサイズを小さくするために誘電率と導電率のどちらか一つを選択する
		writedata(outfile, Ndata, Nx, Ny, Nout, NTx, NRx, Epsr, S_r, S_i);
		//writedata(outfile, Ndata, Nx, Ny, Nout, NTx, NRx, Sigm, S_r, S_i);

		// debug, S行列平均数値出力
		if (log1) {
			log_s1(Ndata, NTx, NRx, S_r, S_i);
		}

		// debug, S行列数値出力
		if (log3) {
			log_s2(Ndata, NTx, NRx, S_r, S_i);
		}

		// debug, 電界分布図形出力
		if (fig2) {
			const int npage = (fig2 < Ndata) ? fig2 : Ndata;
			plot2d_e(8, 4, NTx, 1, Nx, Ny, npage, Ea, 0, 0, iTx, jTx, Nrect, rect, fig2ev2);
		}

		// debug, S行列図形出力
		if (fig3) {
			plot2d_s(1, 4, Ndata, 1, NTx, NRx, S_r, S_i, 0, fig3ev2);
		}
	}

	// 終了時刻
	const double tstop = cputime();
	if (log1 && (commrank == 0)) {
		printf("CPU=%.3f[sec]\n", tstop - tstart);
		fflush(stdout);
	}

	// MPI
	mpi_close();

	return 0;
}
