/*
pml.c
PML-ABC
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

extern void pmlfactor(int lpml, float *f);

extern float C;

// Ez
void pmlEz(int lpml, int Nx, int Ny, float Dt, float Eps0,
	float **Ez, float **Hx, float **Hy, float **Ezx, float **Ezy,
	float *Xn, float *Yn, float *dXn, float *dYn)
{
	float *f = (float *)malloc((2 * lpml + 1) * sizeof(float));
	pmlfactor(lpml, f);

	float *fezx = (float *)malloc((Nx + 1 + 2 * lpml) * sizeof(float));
	float *fezy = (float *)malloc((Ny + 1 + 2 * lpml) * sizeof(float));
	const float cdt = C * Dt;
	float fc = 0;
	for (int i = -lpml; i < Nx + lpml + 1; i++) {
		if      (i <      1) fc = f[- 2 * (i     )    ] / (Xn[1 ] - Xn[0     ]);
		else if (i > Nx - 1) fc = f[+ 2 * (i - Nx)    ] / (Xn[Nx] - Xn[Nx - 1]);
		else                 fc = 0;
		fezx[i + lpml] = 1 / (1 + cdt * fc);
	}
	for (int j = -lpml; j < Ny + lpml + 1; j++) {
		if      (j <      1) fc = f[- 2 * (j     )    ] / (Yn[1 ] - Yn[0     ]);
		else if (j > Ny - 1) fc = f[+ 2 * (j - Ny)    ] / (Yn[Ny] - Yn[Ny - 1]);
		else                 fc = 0;
		fezy[j + lpml] = 1 / (1 + cdt * fc);
	}

	const float r_eps = 1 / Eps0;
	for (int i = -lpml + 1; i < Nx + lpml + 0; i++) {
	for (int j = -lpml + 1; j < Ny + lpml + 0; j++) {
		if ((i < 0) || (Nx - 0 < i) ||
		    (j < 0) || (Ny - 0 < j)) {
			const float dx = dXn[(i < 0) ? 0 : (i > Nx - 0) ? Nx - 0 : i] * r_eps;
			const float dhy = Hy[i     + lpml][j     + lpml]
			                - Hy[i - 1 + lpml][j     + lpml];
			Ezx[i + lpml][j + lpml] = (
			Ezx[i + lpml][j + lpml] + dx * dhy) * fezx[i + lpml];

			const float dy = dYn[(j < 0) ? 0 : (j > Ny - 0) ? Ny - 0 : j] * r_eps;
			const float dhx = Hx[i     + lpml][j     + lpml]
			                - Hx[i     + lpml][j - 1 + lpml];
			Ezy[i + lpml][j + lpml] = (
			Ezy[i + lpml][j + lpml] - dy * dhx) * fezy[j + lpml];

			Ez[i + lpml][j + lpml] = Ezx[i + lpml][j + lpml]
			                       + Ezy[i + lpml][j + lpml];
		}
	}
	}

	free(fezx);
	free(fezy);
	free(f);
}


// Hx
void pmlHx(int lpml, int Nx, int Ny, float Dt,
	float **Ez, float **Hx, float *Yn, float *dYc)
{
	float *f = (float *)malloc((2 * lpml + 1) * sizeof(float));
	pmlfactor(lpml, f);

	float *fhx = (float *)malloc((Ny + 0 + 2 * lpml) * sizeof(float));
	const float cdt = C * Dt;
	float fc = 0;
	for (int j = -lpml; j < Ny + lpml + 0; j++) {
		if      (j <      0) fc = f[- 2 * (j     ) - 1] / (Yn[1 ] - Yn[0     ]);
		else if (j > Ny - 1) fc = f[+ 2 * (j - Ny) + 1] / (Yn[Ny] - Yn[Ny - 1]);
		else                 fc = 0;
		fhx[j + lpml] = 1 / (1 + cdt * fc);
	}

	const float r_amu = 1;  // 磁性体なし
	for (int i = -lpml + 1; i < Nx + lpml + 0; i++) {
	for (int j = -lpml + 0; j < Ny + lpml + 0; j++) {
		if ((i < 0) || (Nx - 0 < i) ||
		    (j < 0) || (Ny - 1 < j)) {
			const float dy = dYc[(j < 0) ? 0 : (j > Ny - 1) ? Ny - 1 : j] * r_amu;
			const float dez = Ez[i    + lpml][j + 1 + lpml]
			                - Ez[i    + lpml][j     + lpml];
			Hx[i + lpml][j + lpml] = (
			Hx[i + lpml][j + lpml] - dy * dez) * fhx[j + lpml];
		}
	}
	}

	free(fhx);
	free(f);
}


// Hy
void pmlHy(int lpml, int Nx, int Ny, float Dt,
	float **Ez, float **Hy, float *Xn, float *dXc)
{
	float *f = (float *)malloc((2 * lpml + 1) * sizeof(float));
	pmlfactor(lpml, f);

	float *fhy = (float *)malloc((Nx + 0 + 2 * lpml) * sizeof(float));
	const float cdt = C * Dt;
	float fc = 0;
	for (int i = -lpml; i < Nx + lpml + 0; i++) {
		if      (i <      0) fc = f[- 2 * (i     ) - 1] / (Xn[1 ] - Xn[0     ]);
		else if (i > Nx - 1) fc = f[+ 2 * (i - Nx) + 1] / (Xn[Nx] - Xn[Nx - 1]);
		else                 fc = 0;
		fhy[i + lpml] = 1 / (1 + cdt * fc);
	}

	const float r_amu = 1;  // 磁性体なし
	for (int i = -lpml + 0; i < Nx + lpml + 0; i++) {
	for (int j = -lpml + 1; j < Ny + lpml + 0; j++) {
		if ((i < 0) || (Nx - 1 < i) ||
		    (j < 0) || (Ny - 0 < j)) {
			const float dx = dXc[(i < 0) ? 0 : (i > Nx - 1) ? Nx - 1 : i] * r_amu;
			const float dez = Ez[i + 1 + lpml][j    + lpml]
			                - Ez[i     + lpml][j    + lpml];
			Hy[i + lpml][j + lpml] = (
			Hy[i + lpml][j + lpml] + dx * dez) * fhy[i + lpml];
		}
	}
	}

	free(fhy);
	free(f);
}
