/*
solve.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>

static void setupFactor(int, int, float, float **, float **, float, float, float **, float **);
static void initfield(int, int, int, int, float **, float **, float **, float *, float *, float **, float **, float **, float **, int, float **, float **);
static void dft(int, int, int, int, int *, int *, float **, float, float, float *, float *, int, float **, float **);
static void average(int, int, int, float **, float **, float **, float []);

extern void dftfactor(int, float, float, float *, float *);
extern void updateEz(int, int, int, float **, float **, float **, float *, float *, float **, float **);
extern void updateHx(int, int, int, float **, float **, float *);
extern void updateHy(int, int, int, float **, float **, float *);
extern void murEz(int, int, float, float *, float *, float **, float **, float **, float);
extern void pmlEz(int, int, int, float, float, float **, float **, float **, float **, float **, float *, float *, float *, float *);
extern void pmlHx(int, int, int, float, float **, float **, float *, float *);
extern void pmlHy(int, int, int, float, float **, float **, float *, float *);
extern float vfeed(float, float);

extern float C;

void solve(
	int Niter, int Nx, int Ny, float Freq, float Dt, int ABC,
	int iTx, int jTx, int NRx, int *iRx, int *jRx,
	float **Ez, float **Hx, float **Hy, float **Ezx, float **Ezy,
	float *Xn, float *Yn, float *dXn, float *dYn, float *dXc, float *dYc,
	float **epsr, float **sigm, float Eps0, float Sig0,
	float **C1Ez, float **C2Ez,
	float **Ezmurx, float **Ezmury,
	float *S_r, float *S_i,
	int ilog, int iout, float **Ez_r, float **Ez_i, float **Ea)
{
	//printf("%d %d %d %e %e %d\n", Niter, Nx, Ny, Freq, Dt, ABC);
	//printf("%d %d %d %d %d\n", iTx, jTx, NRx, jRx[0], iRx[0]);

	// DFT係数
	float *fcos = (float *)malloc((Niter + 1) * sizeof(float));
	float *fsin = (float *)malloc((Niter + 1) * sizeof(float));
	dftfactor(Niter, Freq, Dt, fcos, fsin);

	// FDTD係数
	setupFactor(Nx, Ny, Dt, epsr, sigm, Eps0, Sig0, C1Ez, C2Ez);

	// 電磁界初期化
	initfield(ABC, Nx, Ny, NRx, Ez, Hx, Hy, S_r, S_i, Ezmurx, Ezmury, Ezx, Ezy, iout, Ez_r, Ez_i);

	// FDTD反復計算
	float t = 0;
	for (int iter = 0; iter <= Niter; iter++) {
		// Ez更新
		updateEz(ABC, Nx, Ny, Ez, Hx, Hy, dXn, dYn, C1Ez, C2Ez);

		// 給電点
		Ez[iTx + ABC][jTx + ABC] = vfeed(t, Freq);

		// ABC
		if (ABC > 0) {
			pmlEz(ABC, Nx, Ny, Dt, Eps0, Ez, Hx, Hy, Ezx, Ezy, Xn, Yn, dXn, dYn);
		}
		else {
			murEz(Nx, Ny, Dt, Xn, Yn, Ez, Ezmurx, Ezmury, Eps0);
		}

		// Hx/Hy更新
		updateHx(ABC, Nx, Ny, Ez, Hx, dYc);
		updateHy(ABC, Nx, Ny, Ez, Hy, dXc);

		// ABC
		if (ABC > 0) {
			pmlHx(ABC, Nx, Ny, Dt, Ez, Hx, Yn, dYc);
			pmlHy(ABC, Nx, Ny, Dt, Ez, Hy, Xn, dXc);
		}

		// DFT
		dft(ABC, Nx, Ny, NRx, iRx, jRx, Ez, fcos[iter], fsin[iter], S_r, S_i, iout, Ez_r, Ez_i);

		// 時間ステップ
		t += Dt;

		// debug, 平均電磁界出力
		if (ilog && (iter % 20 == 0)) {
			float fsum[2];
			average(ABC, Nx, Ny, Ez, Hx, Hy, fsum);
			printf("%5d %.6f %.6f\n", iter, fsum[0], fsum[1]);
		}
	}

	// debug, 電界絶対値, セル中心
	if (iout) {
		for (int i = 0; i < Nx; i++) {
		for (int j = 0; j < Ny; j++) {
			// セル平均
			const float e_r = (Ez_r[i][j] + Ez_r[i + 1][j] + Ez_r[i][j + 1] + Ez_r[i + 1][j + 1]) / 4;
			const float e_i = (Ez_i[i][j] + Ez_i[i + 1][j] + Ez_i[i][j + 1] + Ez_i[i + 1][j + 1]) / 4;
			Ea[i][j] = (float)sqrt((e_r * e_r) + (e_i * e_i));
		}
		}
	}

	// free
	free(fcos);
	free(fsin);
}


// 電磁界初期化
static void initfield(int ABC, int Nx, int Ny, int NRx,
	float **Ez, float **Hx, float **Hy,
	float *S_r, float *S_i,
	float **Ezmurx, float **Ezmury, float **Ezx, float **Ezy,
	int iout, float **Ez_r, float **Ez_i)
{
	assert((Nx > 0) && (Ny > 0) && (NRx > 0));

	for (int i = - ABC; i < Nx + 1 + ABC; i++) {
		memset(Ez[i + ABC], 0, (Ny + 1 + 2 * ABC) * sizeof(float));
	}
	for (int i = - ABC; i < Nx + 1 + ABC; i++) {
		memset(Hx[i + ABC], 0, (Ny + 0 + 2 * ABC) * sizeof(float));
	}
	for (int i = - ABC; i < Nx + 0 + ABC; i++) {
		memset(Hy[i + ABC], 0, (Ny + 1 + 2 * ABC) * sizeof(float));
	}

	// S行列
	memset(S_r, 0, NRx * sizeof(float));
	memset(S_i, 0, NRx * sizeof(float));

	// ABC
	if (ABC == 0) {
		for (int m = 0; m < 2; m++) {
			memset(Ezmurx[m], 0, (Ny + 1) * sizeof(float));
			memset(Ezmury[m], 0, (Nx + 1) * sizeof(float));
		}
	}
	else {
		for (int i = - ABC; i < Nx + 1 + ABC; i++) {
			memset(Ezx[i + ABC], 0, (Ny + 1 + 2 * ABC) * sizeof(float));
			memset(Ezy[i + ABC], 0, (Ny + 1 + 2 * ABC) * sizeof(float));
		}
	}

	// debug
	if (iout) {
		for (int i = 0; i < Nx + 1; i++) {
			memset(Ez_r[i], 0, (Ny + 1) * sizeof(float));
			memset(Ez_i[i], 0, (Ny + 1) * sizeof(float));
		}
	}
}


// FDTD係数
static void setupFactor(int Nx, int Ny, float Dt, float **epsr, float **sigm, float Eps0, float Sig0, float **C1Ez, float **C2Ez)
{
	const float pi = (float)(4 * atan(1));
	const float eta = C * 4 * pi * 1e-7f;

	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 1; j++) {
		float eps, sig;
		if ((i == 0) || (i == Nx) || (j == 0) || (j == Ny)) {
			// 境界
			eps = Eps0;
			sig = Sig0;
		}
		else {
			// 内部:4点の平均
			eps = (epsr[i - 1][j - 1] + epsr[i - 1][j] + epsr[i][j - 1] + epsr[i][j]) / 4;
			sig = (sigm[i - 1][j - 1] + sigm[i - 1][j] + sigm[i][j - 1] + sigm[i][j]) / 4;
		}
		C2Ez[i][j] = 1 / (eps + (sig * eta * C * Dt));
		C1Ez[i][j] = C2Ez[i][j] * eps;
	}
	}
}


// 平均電磁界
static void average(int ABC, int Nx, int Ny, float **Ez, float **Hx, float **Hy, float fsum[])
{
	// <E>
	float sume = 0;
	int nume = 0;
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 1; j++) {
		sume += (float)fabs(Ez[i + ABC][j + ABC]);
		nume++;
	}
	}
	fsum[0] = sume/ nume;

	// <H>
	float sumh = 0;
	int numh = 0;
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 0; j++) {
		sumh += (float)fabs(Hx[i + ABC][j + ABC]);
		numh++;
	}
	}
	for (int i = 0; i < Nx + 0; i++) {
	for (int j = 0; j < Ny + 1; j++) {
		sumh += (float)fabs(Hy[i + ABC][j + ABC]);
		numh++;
	}
	}
	fsum[1] = sumh / numh;
}


// DFT
static void dft(int ABC, int Nx, int Ny, int NRx, int *iRx, int *jRx, float **Ez, float fcos, float fsin, float *S_r, float *S_i, int iout, float **Ez_r, float **Ez_i)
{
	// 受信点
	for (int n = 0; n < NRx; n++) {
		const int i = iRx[n];
		const int j = jRx[n];
		S_r[n] += fcos * Ez[i + ABC][j + ABC];
		S_i[n] -= fsin * Ez[i + ABC][j + ABC];
	}

	// debug, 全領域調和界, 計算時間が増えるので必要なときのみ
	if (iout) {
		for (int i = 0; i < Nx + 1; i++) {
		for (int j = 0; j < Ny + 1; j++) {
			Ez_r[i][j] += fcos * Ez[i + ABC][j + ABC];
			Ez_i[i][j] -= fsin * Ez[i + ABC][j + ABC];
		}
		}
	}
}
