/*
utils.c
fdtd2d/fdtd3d共通
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include <assert.h>

extern double urand(int, int64_t *);

// 給電電圧（微分ガウス）
float vfeed(float t, float Freq)
{
	const double tw = 1.27 / Freq;
	const double arg = (t - tw) / (tw / 4);
	const double v = sqrt(2) * exp(0.5) * arg * exp(-arg * arg);

	return (float)v;
}


// DFT係数
void dftfactor(int Niter, float Freq, float Dt, float *fcos, float *fsin)
{
	const double pi = 4 * atan(1);
	const double omega = 2 * pi * Freq;

	// 分母
	double sum_r = 0;
	double sum_i = 0;;
	for (int iter = 0; iter <= Niter; iter++) {
		const float t = iter * Dt;
		const float v = vfeed(t, Freq);
		sum_r += cos(omega * t) * v;
		sum_i -= sin(omega * t) * v;
	}
	const float fnorm = (float)((sum_r * sum_r) + (sum_i * sum_i));

	// DFT係数
	for (int iter = 0; iter <= Niter; iter++) {
		const float t = iter * Dt;
		fcos[iter] = (float)cos(omega * t) / fnorm;
		fsin[iter] = (float)sin(omega * t) / fnorm;
	}
}


// PML減衰係数
void pmlfactor(int lpml, float *f)
{
	// PMLパラメーター
	const float pml_m = 2;
	const float pml_r0 = 1e-8f;

	const float c = (float)((pml_m + 1) / (2.0 * lpml) * log(1.0 / pml_r0));
	//printf("%e\n", c);
	/*f[0] = f[1] = 0;
	for (int n = 2; n < 2 * lpml + 1; n++) {
		f[n] = (float)(c * pow((n - 1) / (2.0 * lpml), pml_m));
	}*/
	f[0] = 0;
	for (int n = 1; n < 2 * lpml + 1; n++) {
		f[n] = (float)(c * pow((n + 0) / (2.0 * lpml), pml_m));
	}
	//for (int n = 0; n < 2 * lpml + 1; n++) printf("%d %f\n", n, f[n]); exit(0);
}


// ランダムに区間を取り出す
// n : 全体の大きさ
// nout : 両端の余白
// lmin, lmax : 長さの最小, 最大
// n0, n1 : 区間の下限, 上限
void getspan(int n, int nout, int lmin, int lmax, int64_t *seed, int *n0, int *n1)
{
	float r1 = (float)urand(3, seed);
	float r2 = (float)urand(3, seed);

	const int ls = (int)(lmin + r1 * (lmax - lmin) + 0.5);
	const int ds = n - 2 * nout - ls;
	const float pc = nout + (0.5f * ls) + (r2 * ds);

	*n0 = (int)(pc - (0.5 * ls) + 0.5);
	*n1 = (int)(pc + (0.5 * ls) + 0.5);

	assert((*n0 >= nout) && (*n0 <= n - nout) && (*n1 >= nout) && (*n1 <= n - nout));
}


// 最小・最大(3D配列)
void getminmax3d(int i0, int i1, int j0, int j1, int k0, int k1, int db, float ***d, float *dmin, float *dmax)
{
	*dmin = +1e12f;
	*dmax = -1e12f;
	for (int i = i0; i < i1; i++) {
	for (int j = j0; j < j1; j++) {
	for (int k = k0; k < k1; k++) {
		const float f = d[i][j][k];
		if (f < *dmin) {
			*dmin = f;
		}
		if (f > *dmax) {
			*dmax = f;
		}
	}
	}
	}

	// dB
	if (db) {
		const float eps = 1e-12f;
		*dmax = 20 * (float)log10(fmax(*dmax, eps));
		*dmin = *dmax - db;
	}
}


// 最小・最大(4D配列)
void getminmax4d(int nn, int nx, int ny, int nz, int db, float ****d, float *dmin, float *dmax)
{
	*dmin = +1e12f;
	*dmax = -1e12f;
	for (int n = 0; n < nn; n++) {
	for (int i = 0; i < nx; i++) {
	for (int j = 0; j < ny; j++) {
	for (int k = 0; k < nz; k++) {
		const float f = d[n][i][j][k];
		if (f < *dmin) {
			*dmin = f;
		}
		if (f > *dmax) {
			*dmax = f;
		}
	}
	}
	}
	}

	// dB
	if (db) {
		const float eps = 1e-12f;
		*dmax = 20 * (float)log10(fmax(*dmax, eps));
		*dmin = *dmax - db;
	}
}


// debug, S行列平均数値出力
void log_s1(int Ndata, int NTx, int NRx, float ***S_r, float ***S_i)
{
	double sum = 0;
	for (int idata = 0; idata < Ndata; idata++) {
	for (int itx = 0; itx < NTx; itx++) {
	for (int irx = 0; irx < NRx; irx++) {
		sum += fabs(S_r[idata][itx][irx])
		     + fabs(S_i[idata][itx][irx]);
	}
	}
	}
	if (Ndata * NTx * NRx > 0) {
		sum /= Ndata * NTx * NRx;
	}

	printf("Save=%.8e\n", sum);
	fflush(stdout);
}


// debug, S行列数値出力, 見やすくするために1000倍
void log_s2(int Ndata, int NTx, int NRx, float ***S_r, float ***S_i)
{
	if (Ndata == 0) return;
	const int idata = 0;  // 最初のデータのみ

	for (int itx = 0; itx < NTx; itx++) {
	for (int irx = 0; irx < NRx; irx++) {
		printf("%4d %4d %10.4f %10.4f\n", itx, irx,
			1000 * S_r[idata][itx][irx],
			1000 * S_i[idata][itx][irx]);
	}
	}
	fflush(stdout);
}


// 出力ファイルサイズ[MB]
int file_size(int Ndata, int Mx, int My, int Mz, int NTx, int NRx)
{
	size_t size = 0;

	size += Ndata * Mx * My * Mz * sizeof(unsigned char);  // Epsr or Sigm
	size += Ndata * 2 * NTx * NRx * sizeof(float);         // S_r, S_i

	return (int)(size * 1e-6) + 1;

}
