/*
FDTD (3D)
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#ifdef _OPENMP
#include <omp.h>
#endif

#include "../fdtd2d/alloc.h"

extern void make_mesh(int, int, int, float, float, float, float, float *, float *, float *, float *, float *, float *, float *, float *, float *, float *);
extern void txrx(int, int, int *, int *, int *, int, int *, int *, int *, int, int *, int *, int *);
extern void geometry(int, int, int64_t, const int [], const float [], const float [], int, int, int, int, int, float ****, float ****, float, float, int **);
extern void solve(int, int, int, int, float, float, int, int, int, int, int, int *, int *, int *, float *, float *, float *, float *, float *, float *, float *, float *, float *, float ***, float ***, float, float, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float *, float *, int, int, float ***, float ***, float ***, float ***, float ***, float ***, float ***);
extern void writedata(const char [], int, int, int, int, int, int, int, float ****, float ***, float ***);
extern void plot2d_g(int, int, int, int, int, int, float ****, int, int, int, int *, int *, int *, int, int *, int *, int *, const char []);
extern void plot2d_e(int, int, int, int, int, int, float ****, int, int, int, int **, const char []);
extern void plot2d_s(int, int, int, int, int, int, float ***, float ***, int, const char []);
extern void plot3d_g(int, int, int, int, int, int, int, int **, int, int *, int *, int *, int, int *, int *, int *, const char []);
extern void log_s1(int, int, int, float ***, float ***);
extern void log_s2(int, int, int, float ***, float ***);
extern int memory_size(int, int, int, int, int, int, int, int, int);
extern int file_size(int, int, int, int, int, int);
extern void mpi_init(int, char *[], int *, int *);
extern void mpi_close(void);
extern void comm_S(int, int, int, int, int, int, float ***, float ***);
extern double cputime(void);

const float C = 2.99792458e8f;  // 光速

int main(int argc, char *argv[])
{
	// 定数
	const int   Ndata = 100;             // データ数
	const int   Nrect = 2;               // 直方体数
	const int   Lrect[] = {3, 11, 1};    // 直方体パラメーター(最小,最大,余白)
	const float Eps[] = {1.5f, 2.5f};    // 誘電率(最小,最大)
	const float Sig[] = {0.02f, 0.08f};  // 導電率[S/m](最小,最大)
	const int   Nsmot = 2;               // 平滑化回数(0以上)
	const int   NTx   = 16;              // 送信点数(16限定)
	const int   NRx   = 16;              // 受信点数(16限定)
	const float Dx    = 0.005f;          // X方向セルサイズ
	const float Dy    = 0.005f;          // Y方向セルサイズ
	const float Dz    = 0.005f;          // Z方向セルサイズ
	const int   Nout  = 4;               // 外部領域セル数
	const int   Lant  = 2;               // アンテナ位置(対象領域境界からの距離)
	const int   ABC   = 6;               // PML層数(0のときはMur一次)
	const float Freq  = 3e9f;            // 周波数
	const int   Niter = 300;             // タイムステップ数
	const float Eps0 = 1;                // 背景媒質比誘電率
	const float Sig0 = 0;                // 背景媒質導電率[S/m]
	const int64_t Seed = 10000;          // 乱数の種, 10000:学習用, 20000:テスト用
	char outfile[BUFSIZ] = "fdtd.bin";   // 出力ファイル名

	// debug
	const int log1 = 1;  // (数値出力) 計算条件とS行列平均(0/1, 通常1)
	const int log2 = 0;  // (数値出力) 収束状況(0/1, 通常0)
	const int log3 = 0;  // (数値出力) S行列(0/1, 通常0)
	const int fig1 = 0;  // (図形出力) 誘電率分布,送受信点(2D/3D)(0/1, 通常0)
	const int fig2 = 0;  // (図形出力) 電界分布(2D, 複数ページ対応)(0,1,2..., 通常0)
	const int fig3 = 0;  // (図形出力) S行列(2D)(0/1, 通常0)
	const char fig1ev2[] = "fig1.ev2";  // $ ev2d.exe fig1.ev2
	const char fig1ev3[] = "fig1.ev3";  // $ ev3d_otk.exe fig1.ev3
	const char fig2ev2[] = "fig2.ev2";  // $ ev2d.exe fig2.ev2
	const char fig3ev2[] = "fig3.ev2";  // $ ev2d.exe fig3.ev2

	// MPI
	int commsize, commrank;
	mpi_init(argc, argv, &commsize, &commrank);
	//printf("%d %d\n", commsize, commrank);

	// 引数: スレッド数、出力ファイル
	int nthread = 1;
	if (argc > 1) {
		nthread = atoi(argv[1]);
	}
	if (argc > 2) {
		strcpy(outfile, argv[2]);
	}
	//printf("%d %s\n", nthread, outfile);

	// OpenMP
#ifdef _OPENMP
	omp_set_num_threads(nthread);
#endif

	assert((NRx == 16) && (NRx == 16));
	assert((Nout > Lant) && (Lant > 0));
	assert(ABC >= 0);

	// 開始時刻
	const double tstart = cputime();

	// 送受信点座標
	int *iTx, *jTx, *kTx, *iRx, *jRx, *kRx;
	alloc1d(int, iTx, NTx)
	alloc1d(int, jTx, NTx)
	alloc1d(int, kTx, NTx)
	alloc1d(int, iRx, NRx)
	alloc1d(int, jRx, NRx)
	alloc1d(int, kRx, NRx)
	int Nx, Ny, Nz;
	txrx(Nout, Lant, &Nx, &Ny, &Nz, NTx, iTx, jTx, kTx, NRx, iRx, jRx, kRx);

	// メッシュ作成
	float Dt;
	float *Xn, *Yn, *Zn, *dXn, *dYn, *dZn, *dXc, *dYc, *dZc;
	alloc1d(float, Xn,  Nx + 1)
	alloc1d(float, Yn,  Ny + 1)
	alloc1d(float, Zn,  Nz + 1)
	alloc1d(float, dXn, Nx + 1)
	alloc1d(float, dYn, Ny + 1)
	alloc1d(float, dZn, Nz + 1)
	alloc1d(float, dXc, Nx + 0)
	alloc1d(float, dYc, Ny + 0)
	alloc1d(float, dZc, Nz + 0)
	make_mesh(Nx, Ny, Nz, Dx, Dy, Dz, Eps0, &Dt, Xn, Yn, Zn, dXn, dYn, dZn, dXc, dYc, dZc);

	// 計算条件表示
	if (log1 && (commrank == 0)) {
		const int mx = Nx - 2 * Nout;
		const int my = Ny - 2 * Nout;
		const int mz = Nz - 2 * Nout;
		const int mem = memory_size(nthread, Ndata, Nx, Ny, Nz, NTx, NRx, ABC, fig2);
		const int filesize = file_size(Ndata, mx, my, mz, NTx, NRx);
		printf("data=%d Nx=%d Ny=%d Nz=%d Mx=%d My=%d Mz=%d Tx=%d Rx=%d rect=%d(%d)\n",
			Ndata, Nx, Ny, Nz, mx, my, mz, NTx, NRx, Nrect, Nsmot);
		printf("process=%d thread=%d memory=%dMB ABC=%d iter=%d output=%s(%dMB)\n",
			commsize, nthread, mem, ABC, Niter, outfile, filesize);
		fflush(stdout);
	}

	// 長方形(debug)
	int **rect = NULL;
	alloc2d(int, rect, Ndata, 6 * Nrect)

	// 誘電率配列(セル中心, xデータ数)
	float ****Epsr, ****Sigm;
	alloc4d(float, Epsr, Ndata, Nx, Ny, Nz)
	alloc4d(float, Sigm, Ndata, Nx, Ny, Nz)

	// 内部領域の誘電率配列に代入
	geometry(Ndata, Nrect, Seed, Lrect, Eps, Sig, Nsmot, Nx, Ny, Nz, Nout, Epsr, Sigm, Eps0, Sig0, rect);

	// debug, 誘電率分布図形出力
	if (((fig1 == 1) || (fig1 == -1)) && (commrank == 0)) {
		plot2d_g(Ndata, 1, Nx, Ny, Nz, Nout, Epsr, 1, 0, NTx, iTx, jTx, kTx, NRx, iRx, jRx, kRx, fig1ev2);
		plot3d_g(Nx, Ny, Nz, Nout, Ndata, 10, Nrect, rect, NTx, iTx, jTx, kTx, NRx, iRx, jRx, kRx, fig1ev3);
	}
	if (fig1 < 0) {
		exit(0);  // debug: 計算せずに終了する
	}

	// debug, 全領域調和界, 図形表示用, Yee格子, PML層を含まない, データ0(=スレッド0)のみ
	float ***Ex_r = NULL, ***Ex_i = NULL;
	float ***Ey_r = NULL, ***Ey_i = NULL;
	float ***Ez_r = NULL, ***Ez_i = NULL;
	if (fig2 && (commrank == 0)) {
		alloc3d(float, Ex_r, Nx + 0, Ny + 1, Nz + 1)
		alloc3d(float, Ex_i, Nx + 0, Ny + 1, Nz + 1)
		alloc3d(float, Ey_r, Nx + 1, Ny + 0, Nz + 1)
		alloc3d(float, Ey_i, Nx + 1, Ny + 0, Nz + 1)
		alloc3d(float, Ez_r, Nx + 1, Ny + 1, Nz + 0)
		alloc3d(float, Ez_i, Nx + 1, Ny + 1, Nz + 0)
	}

	// debug, 全領域調和界, 図形表示用, セル中心. PML層を含まない, 送信点の配列, データ0のみ
	float ****Ea = NULL;
	if (fig2 && (commrank == 0)) {
		// Z面の集合とする, 図形表示のため, ZXYに注意
		alloc4d(float, Ea, NTx, Nz, Nx, Ny)
	}

	// S行列(xデータ数, 複素数)
	float ***S_r = NULL, ***S_i = NULL;
	alloc3d(float, S_r, Ndata, NTx, NRx)
	alloc3d(float, S_i, Ndata, NTx, NRx)

	// プロセスごとのデータ番号の範囲（ブロック分割）
	const int ndata = (Ndata + (commsize - 1)) / commsize;
	int idata0 = commrank * ndata;
	if (idata0 > Ndata - 1) {
		idata0 = Ndata - 1;
	}
	int idata1 = idata0 + ndata;
	if (idata1 > Ndata) {
		idata1 = Ndata;
	}
	//printf("%d %d %d %d\n", commsize, commrank, idata0, idata1); fflush(stdout);

	// OpenMP並列
#ifdef _OPENMP
#pragma omp parallel
#endif
{
	// FDTD係数(スレッド変数, 磁性体は考えないので電界成分のみ)
	float ***C1Ex = NULL, ***C2Ex = NULL, ***C1Ey = NULL, ***C2Ey = NULL, ***C1Ez = NULL, ***C2Ez = NULL;
	alloc3d(float, C1Ex, Nx + 0, Ny + 1, Nz + 1)
	alloc3d(float, C2Ex, Nx + 0, Ny + 1, Nz + 1)
	alloc3d(float, C1Ey, Nx + 1, Ny + 0, Nz + 1)
	alloc3d(float, C2Ey, Nx + 1, Ny + 0, Nz + 1)
	alloc3d(float, C1Ez, Nx + 1, Ny + 1, Nz + 0)
	alloc3d(float, C2Ez, Nx + 1, Ny + 1, Nz + 0)

	// ABC変数(スレッド変数)
	float ***Exmury = NULL, ***Exmurz = NULL;
	float ***Eymurz = NULL, ***Eymurx = NULL;
	float ***Ezmurx = NULL, ***Ezmury = NULL;
	float ***Exy = NULL, ***Exz = NULL;
	float ***Eyz = NULL, ***Eyx = NULL;
	float ***Ezx = NULL, ***Ezy = NULL;
	float ***Hxy = NULL, ***Hxz = NULL;
	float ***Hyz = NULL, ***Hyx = NULL;
	float ***Hzx = NULL, ***Hzy = NULL;
	if (ABC == 0) {
		alloc3d(float, Exmury, 2, Nz + 1, Nx + 0)
		alloc3d(float, Exmurz, 2, Ny + 1, Nx + 0)
		alloc3d(float, Eymurz, 2, Nx + 1, Ny + 0)
		alloc3d(float, Eymurx, 2, Nz + 1, Ny + 0)
		alloc3d(float, Ezmurx, 2, Ny + 1, Nz + 0)
		alloc3d(float, Ezmury, 2, Nx + 1, Nz + 0)
	}
	else {
		alloc3d(float, Exy, Nx + 0 + 2 * ABC, Ny + 1 + 2 * ABC, Nz + 1 + 2 * ABC)
		alloc3d(float, Exz, Nx + 0 + 2 * ABC, Ny + 1 + 2 * ABC, Nz + 1 + 2 * ABC)
		alloc3d(float, Eyz, Nx + 1 + 2 * ABC, Ny + 0 + 2 * ABC, Nz + 1 + 2 * ABC)
		alloc3d(float, Eyx, Nx + 1 + 2 * ABC, Ny + 0 + 2 * ABC, Nz + 1 + 2 * ABC)
		alloc3d(float, Ezx, Nx + 1 + 2 * ABC, Ny + 1 + 2 * ABC, Nz + 0 + 2 * ABC)
		alloc3d(float, Ezy, Nx + 1 + 2 * ABC, Ny + 1 + 2 * ABC, Nz + 0 + 2 * ABC)
		alloc3d(float, Hxy, Nx + 1 + 2 * ABC, Ny + 0 + 2 * ABC, Nz + 0 + 2 * ABC)
		alloc3d(float, Hxz, Nx + 1 + 2 * ABC, Ny + 0 + 2 * ABC, Nz + 0 + 2 * ABC)
		alloc3d(float, Hyz, Nx + 0 + 2 * ABC, Ny + 1 + 2 * ABC, Nz + 0 + 2 * ABC)
		alloc3d(float, Hyx, Nx + 0 + 2 * ABC, Ny + 1 + 2 * ABC, Nz + 0 + 2 * ABC)
		alloc3d(float, Hzx, Nx + 0 + 2 * ABC, Ny + 0 + 2 * ABC, Nz + 1 + 2 * ABC)
		alloc3d(float, Hzy, Nx + 0 + 2 * ABC, Ny + 0 + 2 * ABC, Nz + 1 + 2 * ABC)
	}

	// 電磁界配列(スレッド変数)
	float ***Ex, ***Ey, ***Ez, ***Hx, ***Hy, ***Hz;
	alloc3d(float, Ex, Nx + 0 + 2 * ABC, Ny + 1 + 2 * ABC, Nz + 1 + 2 * ABC)
	alloc3d(float, Ey, Nx + 1 + 2 * ABC, Ny + 0 + 2 * ABC, Nz + 1 + 2 * ABC)
	alloc3d(float, Ez, Nx + 1 + 2 * ABC, Ny + 1 + 2 * ABC, Nz + 0 + 2 * ABC)
	alloc3d(float, Hx, Nx + 1 + 2 * ABC, Ny + 0 + 2 * ABC, Nz + 0 + 2 * ABC)
	alloc3d(float, Hy, Nx + 0 + 2 * ABC, Ny + 1 + 2 * ABC, Nz + 0 + 2 * ABC)
	alloc3d(float, Hz, Nx + 0 + 2 * ABC, Ny + 0 + 2 * ABC, Nz + 1 + 2 * ABC)

	// スレッド番号
	const int ithread = omp_get_thread_num();

	// データに関するループ
	//for (int idata = ithread; idata < Ndata; idata += nthread) {
	for (int idata = idata0 + ithread; idata < idata1; idata += nthread) {
		//printf("%d %d %d\n", commrank, ithread, idata); fflush(stdout);
		const int ilog = log2 && (commrank == 0) && (idata == 0);  // 収束状況(データ0のみ)
		const int iout = fig2 && (commrank == 0) && (idata == 0);  // 電界分布(データ0のみ)
		if (log1 && (commrank == 0) && (idata % 1000 == 0)) {printf("%d %d\n", idata, ithread); fflush(stdout);}  // 経過出力
		// 送信点に関するループ
		for (int itx = 0; itx < NTx; itx++) {
			solve(
				Niter, Nx, Ny, Nz, Freq, Dt, ABC,
				iTx[itx], jTx[itx], kTx[itx], NRx, iRx, jRx, kRx,
				Xn, Yn, Zn, dXn, dYn, dZn, dXc, dYc, dZc,
				Epsr[idata], Sigm[idata], Eps0, Sig0,
				Ex, Ey, Ez, Hx, Hy, Hz,
				C1Ex, C2Ex, C1Ey, C2Ey, C1Ez, C2Ez,
				Exmury, Exmurz, Eymurz, Eymurx, Ezmurx, Ezmury,
				Exy, Exz, Eyz, Eyx, Ezx, Ezy,
				Hxy, Hxz, Hyz, Hyx, Hzx, Hzy,
				S_r[idata][itx], S_i[idata][itx],
				ilog, iout, Ex_r, Ex_i, Ey_r, Ey_i, Ez_r, Ez_i, (iout ? Ea[itx] : NULL));
		}
	}
}

	// S行列をrootに集める
	if (commsize > 1) {
		comm_S(commsize, commrank, idata0, idata1, NTx, NRx, S_r, S_i);
	}

	// 出力:root
	if (commrank == 0) {
		// ファイル出力
		// ファイルサイズを小さくするために誘電率と導電率のどちらか一つを選択する
		writedata(outfile, Ndata, Nx, Ny, Nz, Nout, NTx, NRx, Epsr, S_r, S_i);
		//writedata(outfile, Ndata, Nx, Ny, Nz, Nout, NTx, NRx, Sigm, S_r, S_i);

		// debug, S行列平均数値出力
		if (log1) {
			log_s1(Ndata, NTx, NRx, S_r, S_i);
		}

		// debug, S行列数値出力
		if (log3) {
			log_s2(Ndata, NTx, NRx, S_r, S_i);
		}

		// debug, 電界分布図形出力
		if (fig2) {
			plot2d_e(NTx, 1, Nx, Ny, Nz, Nout, Ea, 70, 0, Nrect, rect, fig2ev2);
		}

		// debug, S行列図形出力
		if (fig3) {
			plot2d_s(1, 1, Ndata, 1, NTx, NRx, S_r, S_i, 0, fig3ev2);
		}
	}

	// 終了時刻
	const double tstop = cputime();
	if (log1 && (commrank == 0)) {
		printf("CPU=%.3f[sec]\n", tstop - tstart);
		fflush(stdout);
	}

	// MPI
	mpi_close();

	return 0;
}
